package ch.hslu.dmg.g06.logic;

import java.sql.Date;
import java.util.ArrayList;
import java.util.List;
import java.util.Calendar;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Query;

import ch.hslu.dmg.g06.common.DateUtils;
import ch.hslu.dmg.g06.data.*;

public class Production {
	
	private EntityManager entityManager = null;
	private List<ProductionStep> currentSteps = new ArrayList<ProductionStep>();

	
	public Production(EntityManagerFactory entityFactory) {
				
		entityManager = entityFactory.createEntityManager();
	}

	
	
	public synchronized Date Produce(Order newOrder) {
			
		if (newOrder.getPart() == null) {
			
			System.out.println("Der Bestellung wurde kein Teil zugewiesen");
			return null;
		}

		// Perform a simple query for the part entity
		Query partQuery = entityManager.createNamedQuery("Part.getPartById");
		partQuery.setParameter("id", newOrder.getPart().getId());
		
		Part part  = (Part)partQuery.getResultList().get(0);
		
		// Check if the part exists
		if (part == null) {
			
			System.out.println("Das zu bestellende Teil existiert nicht!");
			return null;
		}
		
		// Check if the part is composed of other parts
		if (part.getSupplier() != null) {
			
			System.out.println("Das " + part + " ist direkt ab Lager verfgbar.");
			return new Date((new java.util.Date()).getTime());
		}	
		
		// Clear last steps
		currentSteps.clear();
		
		// Get Calendar object set to the date and time of the given Date object  
		Calendar calendar = Calendar.getInstance();  
		calendar.setTime(new java.util.Date());  
		  
		// Set time fields to zero  
		calendar.set(Calendar.HOUR_OF_DAY, 0);  
		calendar.set(Calendar.MINUTE, 0);  
		calendar.set(Calendar.SECOND, 0);  
		calendar.set(Calendar.MILLISECOND, 0);
		
		Date startDate = new Date(calendar.getTime().getTime());
		
		
		try {
			Date endDate = calculateProductionDates(part, startDate, 0);
				
		    newOrder.setPart(part);
		    
		    newOrder.setDeliveryDate(endDate);
		    newOrder.setDate(new Date((new java.util.Date().getTime())));
		    
		    entityManager.getTransaction().begin();
		    
		    entityManager.persist(newOrder);
		    for(ProductionStep step : currentSteps) {
		    	entityManager.persist(step);
		    }
		    
		    entityManager.flush();
		    
		    entityManager.getTransaction().commit();
		    		    
		    System.out.print(newOrder + " wird ausgeliefert am " + DateUtils.formatDate(endDate));
			
			return endDate;
		} catch(Exception ex) {
			
			entityManager.getTransaction().rollback();
			
			System.out.println(part + " kann nicht produziert werden: " + ex.getMessage());
			ex.printStackTrace();
			
			return null;
		}		
	}
	
	
	
	private Date calculateProductionDates(Part part, Date startDate, int deepth) throws Exception {
				
		// Only production parts are relevant
		if (part.getMachine() == null) return null;
		
		Date maxDate = startDate;
		for(Part subPart : part.getComposedOf()) {
			
			Date productionDate = calculateProductionDates(subPart, new Date(startDate.getTime()), deepth + 1);
					
			if (productionDate != null && productionDate.getTime() > maxDate.getTime()) {
				maxDate = productionDate;
			}				
		}
		
		// Add one day
		DateUtils.addDay(maxDate);
		
		while(true) {
			
			if (isWorkday(maxDate)) {
				
				if (machineIsAvailable(part.getMachine(), maxDate)) {
				
					Employee availableEmployee = getAvailableEmployee(part.getMachine(), maxDate);
					
					if (availableEmployee != null) {
																		
						System.out.println(getDeepthString(deepth) + part.getMachine() + " und " + availableEmployee + 
										   " werden am Tag " + DateUtils.formatDate(maxDate) + " reserviert, um das " + part + " zu produzieren.");
						
						ProductionStep step = new ProductionStep();
						step.setMachine(part.getMachine());
						step.setEmployee(availableEmployee);
						step.setPart(part);
						step.setDate(new Date(maxDate.getTime()));
						
						currentSteps.add(step);
						
						break;
					}
				} else {
					System.out.println(part.getMachine() + " ist nicht verfgbar am " + DateUtils.formatDate(maxDate));
				}
				
			}			
			
			// Add one day
			DateUtils.addDay(maxDate);
		}
		
		
		return maxDate;
	}
	
	private boolean isWorkday(Date date) {
		
		Calendar calendar = Calendar.getInstance();  		
        calendar.setTime(date);  

        // Get the weekday (Sun = 1)
        int weekday = calendar.get(Calendar.DAY_OF_WEEK);
		
		return weekday >= 2 && weekday <= 6;
	}
	
	private boolean machineIsAvailable(Machine machine, Date date) {
		
		// Check temporary order table
		for (ProductionStep step : currentSteps) {
			
			if (step.getDate().getTime() == date.getTime() &&
				step.getMachine().getId() == machine.getId()) return false;
		}
		
		// Check if the machine is already used on this specific date
		Query partQuery = entityManager.createNamedQuery("ProductionStep.machineIsInUse");
		partQuery.setParameter("id", machine.getId());
		partQuery.setParameter("date", date);
				
		return partQuery.getResultList().size() == 0;
	}
	
	private Employee getAvailableEmployee(Machine machine, Date date) throws Exception {
		
		// Get all users which have the permission to use this machine
		Query partQuery = entityManager.createNamedQuery("CanUse.getPossibleEmployees");
		partQuery.setParameter("id", machine.getId());

		@SuppressWarnings("unchecked")
		List<CanUse> possibleEmployees = (List<CanUse>)partQuery.getResultList();
		
		if (possibleEmployees.size() == 0) {
			throw new Exception("Produktion nicht mglich, da niemand die " + machine + " bedienen kann");
		}
		
		// Check if there is a free employee
		for (CanUse canUse : possibleEmployees) {
			
			if (employeeIsAvailable(canUse.getEmployee(), date)) {
				
				return canUse.getEmployee();
			}
		}
		
		// No free employees found in the database, return null
		return null;
	}
	
	private boolean employeeIsAvailable(Employee employee, Date date) {
		
		// Check temporary order table
		for (ProductionStep step : currentSteps) {
			
			if (employee.getId() == step.getEmployee().getId() &&
				step.getDate().getTime() == date.getTime()) return false;
		}
		
		// Check production steps from the database
		Query partQuery = entityManager.createNamedQuery("ProductionStep.employeeIsBusy");
		partQuery.setParameter("id", employee.getId());
		partQuery.setParameter("date", date);
				
		return partQuery.getResultList().size() == 0;
	}	
	
	private String getDeepthString(int deepth) {
		
		String text = "";
		
		for (int index = 0; index < deepth; index++) text += ""; 
		
		return text + " ";
	}
}
